"""Read all head time series data and store in database. 

Determine selection of time series for modeling based on selection criteria:
   - at least 6 years of data in training period
   - at least 1 year of dat in testing period
   - at least 6 years with 180 unique days with observations

"""
# %%
import json
import logging
import os
import sys

import hydropandas as hpd
import numpy as np
import pandas as pd
import pastas as ps
from tqdm import tqdm

sys.path.insert(1, "..")
from util import get_pastastore


logging.basicConfig(level=logging.INFO)

# %% script settings
update_w_dino = True  # update series with historical dino data
datadir = "../../data"

# %% setup pastastore
with open("../tsa_settings.json", "r") as f:
    settings = json.load(f)

# connect to database
name = f"frontiers_{settings['freq']}"
pstore = get_pastastore(name, settings["connector_type"])
conn = pstore.conn

# %% read csv
oc = hpd.ObsCollection.from_artdino_dir(os.path.join(datadir,
                                                     "piezometers",
                                                     "csv.zip"),
                                        ObsClass=hpd.GroundwaterObs,
                                        subdir="csv",
                                        suffix=".csv",
                                        keep_all_obs=True)

# %% get piezometer layer information from REGIS
try:
    # layer info
    rlay = pd.read_csv(
        "../../data/piezometers/piezometer_layer_information.csv",
        index_col=[0])
except FileNotFoundError:
    mapping = {
        'hlc': "phreatic",
        'bxz2': "aq1",
        'bxz3': "aq1",
        'bxk2': "aq1",
        'bxz4': "aq1",
        'krz2': "aq1",
        'krz3': "aq1",
        'krzuk1': "aq1",
        'krz4': "aq1",
        'krtwk1': "aq1",
        'krz5': "aq1",
        'drz1': "aq1",
        'druik1': "aq1",
        'drz2': "aq1",
        'drgik1': "aq1",
        'drz3': "aq1",
        'dtc': "aq1",
        'pzwaz1': "aq1",
        'wak1': "aq1",
        'pzwaz2': "aq2",
        'wak2': "aq2",
        'pzwaz3': "aq2",
        'wak3': "aq2",
        'pzwaz4': "aq2",
        'msz1': "aq2",
        'msk1': "aq2",
        'msz2': "aq2",
        'msk2': "aq2",
        'msz3': "aq2",
        'msz4': "aq2",
        'ooz1': "aq2",
        'ook1': "aq2",
        'ooz2': "aq2",
        'ooc': "aq2",
        'brz1': "aq2",
        'brk1': "aq2",
        "nan": "nan"
    }

    # overwrite NaN for bottom filter level specific observation well
    oc._set_metadata_value(
        "B39H0383-001", "onderkant_filter", -1.0, add_to_meta=True)

    rlay = oc.gwobs.get_regis_layers()
    rlay = rlay.to_frame()
    rlay["geolayer"] = rlay.applymap(lambda k: mapping[k.lower()])
    rlay.columns = ["regis_layer", "geolayer"]
    rlay.to_csv("../../data/piezometers/piezometer_layer_information.csv")


# %% add historical data from DINOLoket
linktable = pd.read_excel(os.path.join(datadir,
                                       "piezometers",
                                       "overzicht_vitens.xlsx"),
                          engine='openpyxl',
                          index_col=[0])
linktable.replace("onbekend", np.nan, inplace=True)

ocd = hpd.ObsCollection.from_dino(dirname=os.path.join(datadir,
                                                       "piezometers",
                                                       "20211102_dino.zip"))

no_dino_data = []

obs = oc.loc[:, "obs"]
for o in tqdm(obs.values, desc="Add oseries"):

    name = (o.meta["locatie"] + "-" +
            str(o.meta["filternr"]))
    dinoid = (o.meta["locatie"] +
              str(o.meta["filternr"]).zfill(3))

    # one loc w different naming scheme
    if dinoid not in ["P39F0006"]:
        dinoloc = dinoid[:-3]
    else:
        dinoloc = dinoid

    # update with historical data?
    if update_w_dino:
        try:
            # historic timeseries
            odino = ocd.loc[dinoid, "obs"]

            # get mean value per day
            # (only day resolution is returned by DINOLoket)
            dino_series = odino.groupby(by=odino.index).mean()

            # rename column
            dino_series.columns = [icol.lower()
                                   for icol in dino_series.columns]
            # get only data from before first obs in csv file
            dino_series = dino_series.loc[:o.index[0] - pd.Timedelta(days=1)]

            # get only data after tmin
            dino_series = dino_series.loc[~(
                dino_series.index < pd.Timestamp("2000-01-01"))]

            # update series
            s = pd.concat([dino_series, o], axis=0, sort=True)

            # ensure dtypes are clearly defined so it is not stored as Blob
            s["filternummer"] = s["filternummer"].bfill().astype(int)
            s["locatie"] = s["locatie"].bfill().astype(str)

        except KeyError:
            # print("No historical DINO data!")
            no_dino_data += [dinoid]
            s = o
            s["filternummer"] = s["filternummer"].bfill().astype(int)
            s["locatie"] = s["locatie"].bfill().astype(str)

        except Exception as e:
            print(e)
            s = o
    else:
        s = o

    s = s[["stand_m_tov_nap"]]
    s.index = pd.to_datetime(s.index, format='%Y/%m/%d %H%M%S')

    # sample data with specified timestep in settings
    if settings["freq"] != "D":
        dt = pd.Timedelta(settings["oseries_timedelta"])
        new_idx = pd.date_range(s.index[0].floor(settings["freq"] + dt),
                                s.index[-1], freq=settings["freq"])
        sample_idx = ps.utils.get_sample(s.index, new_idx)
        s = s.loc[sample_idx]

    s = s.loc[~s.index.duplicated(keep="last")]

    # add to database
    meta = o.meta
    meta["regis_layer"] = rlay.loc[o.name, "regis_layer"]
    meta["aquifer"] = rlay.loc[o.name, "geolayer"]

    conn.add_oseries(s["stand_m_tov_nap"], o.name, metadata=meta,
                     overwrite=True)

# %% remove duplicated timestamps in timeseries

conn.add_oseries(
    conn.get_oseries(
        'B40C0551-003')[~conn.get_oseries('B40C0551-003').duplicated()],
    'B40C0551-003',
    metadata=conn.oseries.loc['B40C0551-003'].to_dict(),
    overwrite=True)

conn.add_oseries(
    conn.get_oseries(
        'B39F0743-003')[~conn.get_oseries('B39F0743-003').duplicated()],
    'B39F0743-003',
    metadata=conn.oseries.loc['B39F0743-003'].to_dict(),
    overwrite=True)

# %% oseries selection

# split data into train/test set
train_start = "2000-01-01"
train_end = "2014-12-31"

test_start = "2015-01-01"
test_end = "2021-06-30"

# 1. find oseries spanning at least 6 yrs in training period
oseries_tmin_tmax = oc.stats.get_first_last_obs_date()

tdelta = pd.Timestamp(train_end) - oseries_tmin_tmax["date_first_measurement"]
mask1 = tdelta > pd.Timedelta(days=6 * 365)

print(f"Dropped {(~mask1).sum()} oseries because "
      "insufficient data in training period!")

# 2. find oseries spanning at least 1 yrs in testing period
tdelta2 = oseries_tmin_tmax["date_last_measurement"] - pd.Timestamp(test_start)
mask2 = tdelta2 > pd.Timedelta(days=365)

print(f"Dropped {(~mask2).sum()} oseries because "
      "insufficient data in testing period!")

# 3. ensure sufficient unique days with measurements per year
unique_obs_days = []
for o in tqdm(oc.loc[:, "obs"], desc="Get unique obs days"):
    udays = o.index.normalize().unique()
    udays_per_year = udays.to_series().groupby(udays.year).count()
    udays_per_year.name = o.name
    unique_obs_days.append(udays_per_year)

unique_obs_days = pd.concat(unique_obs_days, axis=1)

# 6 years with at least 180 unique measurement days since 2000
mask3 = (unique_obs_days.loc[2000:] > 180).sum() > 6

print(f"Dropped {(~mask3).sum()} oseries because "
      "insufficient unique measurement days per year!")

# get qualifying oseries
model_names = oseries_tmin_tmax.loc[mask1 & mask2 & mask3].index

# %% manually remove timeseries from selection (based on visual inspection)

odf = pstore.oseries.copy().drop(columns=["filename", "metadata_available"])
odf = odf.join(oseries_tmin_tmax)
odf["selected"] = pd.NA
odf["comment"] = ""
odf.loc[model_names, "selected"] = model_names

remove = {
    "B39F0745-003": "remove, funky data, not reliable (email F. v. Vliet 17/11/2021)",
    "B40A0214-001": "remove, near wellfield",
    "B40A0214-002": "remove, near wellfield",
    "B40A0219-001": "remove, near wellfield",
    "B40A0400-003": "remove, near wellfield",
    "B40A0400-005": "remove, near wellfield",
    "B40A0535-002": "remove, near wellfield",
    "B40A2160-003": "remove, near wellfield",
    "B40A2161-003": "remove, near wellfield"
}

for k, v in remove.items():
    odf.loc[k, "selected"] = pd.NA
    odf.loc[k, "comment"] = v

print(f"Dropped {len(remove)} oseries because "
      "of unreliable data or series impossible to model "
      "due to proximity to wellfield!")

# %% write oseries metadata and selection to file
odf.to_csv("../../data/piezometers/oseries_metadata_and_selection.csv")
