# %%
import json
import logging
import os

import hydropandas as hpd
import numpy as np
import pandas as pd
import pastas as ps
import pastastore as pst
from tqdm import tqdm

logging.basicConfig(level=logging.INFO)

# %% script settings
update_w_dino = True  # update series with historical dino data
datadir = "../../data"

# %% setup pastastore
with open("../tsa_settings.json", "r") as f:
    settings = json.load(f)

if settings["freq"] != "D":
    name = f"frontiers_{settings['freq']}"
else:
    name = "frontiers"

connstr = 'mongodb://localhost:27017/'
conn = pst.ArcticConnector(name, connstr)

# %% read csv
oc = hpd.ObsCollection.from_artdino_dir(os.path.join(datadir,
                                                     "piezometers",
                                                     "csv.zip"),
                                        ObsClass=hpd.GroundwaterObs,
                                        subdir="csv",
                                        suffix=".csv",
                                        keep_all_obs=True)

# %% get piezometer layer information
try:
    # layer info
    rlay = pd.read_csv(
        "../../data/piezometers/piezometer_layer_information.csv",
        index_col=[0])
except FileNotFoundError:
    mapping = {
        'hlc': "phreatic",
        'bxz2': "aq1",
        'bxz3': "aq1",
        'bxk2': "aq1",
        'bxz4': "aq1",
        'krz2': "aq1",
        'krz3': "aq1",
        'krzuk1': "aq1",
        'krz4': "aq1",
        'krtwk1': "aq1",
        'krz5': "aq1",
        'drz1': "aq1",
        'druik1': "aq1",
        'drz2': "aq1",
        'drgik1': "aq1",
        'drz3': "aq1",
        'dtc': "aq1",
        'pzwaz1': "aq1",
        'wak1': "aq1",
        'pzwaz2': "aq2",
        'wak2': "aq2",
        'pzwaz3': "aq2",
        'wak3': "aq2",
        'pzwaz4': "aq2",
        'msz1': "aq2",
        'msk1': "aq2",
        'msz2': "aq2",
        'msk2': "aq2",
        'msz3': "aq2",
        'msz4': "aq2",
        'ooz1': "aq2",
        'ook1': "aq2",
        'ooz2': "aq2",
        'ooc': "aq2",
        'brz1': "aq2",
        'brk1': "aq2",
        "nan": "nan"
    }
    rlay = oc.gwobs.get_regis_layer()
    rlay = rlay.to_frame()
    rlay["geolayer"] = rlay.applymap(lambda k: mapping[k.lower()])
    rlay.columns = ["regis_layer", "geolayer"]
    rlay.to_csv("../../data/piezometers/piezometer_layer_information.csv")

# %% add historical data from DINO
linktable = pd.read_excel(os.path.join(datadir,
                                       "piezometers",
                                       "overzicht_vitens.xlsx"),
                          engine='openpyxl',
                          index_col=[0])
linktable.replace("onbekend", np.nan, inplace=True)

ocd = hpd.ObsCollection.from_dino(dirname=os.path.join(datadir,
                                                       "piezometers",
                                                       "20211102_dino.zip"))

no_dino_data = []

obs = oc.loc[:, "obs"]
for o in tqdm(obs.values, desc="Add oseries"):

    name = (o.meta["locatie"] + "-" +
            str(o.meta["filternr"]))
    dinoid = (o.meta["locatie"] +
              str(o.meta["filternr"]).zfill(3))

    if dinoid not in ["P39F0006"]:
        dinoloc = dinoid[:-3]
    else:
        dinoloc = dinoid

    if update_w_dino:
        try:
            # historische dinoreeks
            odino = ocd.loc[dinoid, "obs"]

            # get mean value per day
            dino_series = odino.groupby(by=odino.index).mean()

            # rename column
            dino_series.columns = [icol.lower()
                                   for icol in dino_series.columns]
            # get only data from before first obs in csv file
            dino_series = dino_series.loc[:o.index[0] - pd.Timedelta(days=1)]

            # get only data after tmin
            dino_series = dino_series.loc[~(
                dino_series.index < pd.Timestamp("2000-01-01"))]

            # update series
            s = pd.concat([dino_series, o], axis=0, sort=True)

            # ensure dtypes are clearly defined so it is not stored as Blob
            s["filternummer"] = s["filternummer"].bfill().astype(int)
            s["locatie"] = s["locatie"].bfill().astype(str)

        except KeyError:
            # print("No historical DINO data!")
            no_dino_data += [dinoid]
            s = o
            s["filternummer"] = s["filternummer"].bfill().astype(int)
            s["locatie"] = s["locatie"].bfill().astype(str)

        except Exception as e:
            print(e)
            s = o
    else:
        s = o

    s = s[["stand_m_tov_nap"]]
    s.index = pd.to_datetime(s.index, format='%Y/%m/%d %H%M%S')

    if settings["freq"] != "D":
        new_idx = pd.date_range(s.index[0].floor(settings["freq"]),
                                s.index[-1], freq=settings["freq"])
        sample_idx = ps.utils.get_sample(s.index, new_idx)
        s = s.loc[sample_idx]

    # add to db
    meta = o.meta
    meta["regis_layer"] = rlay.loc[o.name, "regis_layer"]
    meta["aquifer"] = rlay.loc[o.name, "geolayer"]

    conn.add_oseries(s["stand_m_tov_nap"], o.name, metadata=meta,
                     overwrite=True)

# remove duplicated timestamps in timeseries
conn.add_oseries(conn.get_oseries('B40C0551-003')[~conn.get_oseries('B40C0551-003').duplicated()], 'B40C0551-003', metadata=conn.oseries.loc['B40C0551-003'].to_dict(), overwrite=True)
conn.add_oseries(conn.get_oseries('B39F0743-003')[~conn.get_oseries('B39F0743-003').duplicated()], 'B39F0743-003', metadata=conn.oseries.loc['B39F0743-003'].to_dict(), overwrite=True)