"""Read well time series and metadata from Vitens datasets.

Note: well fields consist of multiple pumping wells. In this study, the 
average location of the well fields is used to situate the time series
of the extraction rate. 
"""
# %%
import json
import os
import sys

import pandas as pd
from tqdm import tqdm

sys.path.insert(1, "..")
from util import get_pastastore

# %% setup pastastore
with open("../tsa_settings.json", "r") as f:
    settings = json.load(f)

name = f"frontiers_{settings['freq']}"
pstore = get_pastastore(name, settings["connector_type"])
conn = pstore.conn

# %% Load metadata
datadir = "../../data"
pgr = pd.read_csv(os.path.join(
    datadir, "wells/well_metadata_public.csv"), index_col=[0])
pgr["z"] = pgr.loc[:, ['screen_top_nap', 'screen_bot_nap']].mean(axis=1)

# %% Load timeseries
wells_timeseries_excel = "verzameld_overbetuwe.xlsx"
excelf = os.path.join(datadir, "wells", wells_timeseries_excel)

# monthly data
df_monthly = pd.read_excel(excelf, sheet_name="PGdata_perMaand_1958-2006",
                           parse_dates=[0], index_col=[0], skiprows=1,
                           na_values=["-"], usecols=range(5), engine='openpyxl')
df_monthly = df_monthly.dropna(how="all", axis=0)

# daily data
df_daily = pd.read_excel(excelf, sheet_name="PIdata_perDag_2007-nu",
                         parse_dates=[0], index_col=[0], skiprows=1,
                         na_values=["-"], usecols=range(5), engine='openpyxl')

# loop over columns
for icol in tqdm(df_monthly.columns):

    # get series
    s_monthly = df_monthly[icol]
    s_daily = df_daily[icol]

    # remove NaNs from beginning
    s_daily = s_daily.loc[s_daily.first_valid_index():]
    s_monthly = s_monthly.loc[
        s_monthly.first_valid_index():s_daily.index[0] - pd.Timedelta(days=29)]

    # resample monthly to daily data
    enddate = s_monthly.index[-1] + pd.Timedelta(days=31)
    enddate = enddate - pd.to_timedelta(enddate.day - 1)
    startdate = s_monthly.index[0]
    ind_daily = pd.date_range(
        startdate, enddate - pd.Timedelta(days=1), freq="D")
    s_monthly_to_daily = s_monthly.reindex(
        ind_daily).ffill() / ind_daily.days_in_month

    # concat
    s = pd.concat([s_monthly_to_daily, s_daily], axis=0)
    s.name = icol

    # fix registration error (there are probably more but this was a major one)
    if icol in ["Fikkersdries", "Hemmen", "Zetten"]:
        val = s.loc["2013-06-10"]
        s.loc["2013-06-08":"2013-06-10"] = val / 3

    # metadata
    wmeta = {
        # "x": pgr.loc[icol, "x"],
        # "y": pgr.loc[icol, "y"],
        "z": pgr.loc[icol, "z"],
        "name": icol,
        "units": "Mm$^3/yr",
        # "note" : "xy-coordinates are approximate"
    }

    # add to database, convert units to Mm3/yr
    conn.add_stress(s.to_frame() * 365. * 1e-6, icol, kind="well",
                    metadata=wmeta, overwrite=True)
