"""Calculate interpolated river stage time series at the point on the 
river closest to an observation well. For each observation well a unique
river stage time series is stored.
"""
# %%
import json
import os
import sys

import geopandas as gpd
import numpy as np
import pandas as pd
from pyproj import Proj
from shapely import geometry, ops
from tqdm import tqdm

sys.path.insert(1, "..")
from util import get_pastastore

# %% setup pastastore
with open("../tsa_settings.json", "r") as f:
    settings = json.load(f)

name = f"frontiers_{settings['freq']}"
pstore = get_pastastore(name, settings["connector_type"])
conn = pstore.conn

# %% Shapefiles
datadir = "../../data"
gisdir = os.path.join(datadir, "gis")
rivs = gpd.read_file(os.path.join(gisdir, "main_rivs.shp"))

# combine shapes to form 2 rivers
waal = rivs.loc[0, "geometry"]
nederrijn = ops.unary_union(rivs.loc[1:, "geometry"])

# get obs station locations
wlocs = [
    'Arnhem',
    'Dodewaard',
    'Driel beneden',
    'Driel boven',
    'Grebbe',
    'IJsselkop',
    'Nijmegen haven',
    'Pannerdense Kop'
]
locs = conn.stresses.loc[conn.stresses.kind == "wlvl"].loc[wlocs]
locs["geometry"] = locs.apply(lambda r: geometry.Point(r.x, r.y), axis=1)
locs = gpd.GeoDataFrame(locs)
locs.crs = Proj("epsg:28992").definition_string()

# %% Link locations to river and calculate linear positions

# calculate shortest distance to river from points
waaldist = locs.distance(waal)
nrijndist = locs.distance(nederrijn)
dist = pd.concat([waaldist, nrijndist], axis=1)
dist.columns = ["Waal", "Nederrijn"]

# assign locations to a river (one is on both)
locs["on_waal"] = dist.idxmin(axis=1) == "Waal"
locs["on_nederrijn"] = dist.idxmin(axis=1) == "Nederrijn"
locs.loc["Pannerdense Kop", ["on_waal", "on_nederrijn"]] = True

# calculate distance along shape
locs["waal_dist"] = [waal.project(igeom) for igeom in locs.geometry]
locs.loc[~locs.loc[:, "on_waal"], "waal_dist"] = np.nan
locs["nederrijn_dist"] = [nederrijn.project(igeom) for igeom in locs.geometry]
locs.loc[~locs.loc[:, "on_nederrijn"], "nederrijn_dist"] = np.nan

# %% Calculate interpolation weights

# 0. get oseries as geodataframe
oseries = conn.oseries.copy()
oseries["geometry"] = oseries.apply(
    lambda r: geometry.Point(r.x, r.y), axis=1)
oseries = gpd.GeoDataFrame(oseries)
oseries.crs = Proj("epsg:28992").definition_string()

# 1. link one river to each point
waal_dist = oseries.distance(waal)
waal_dist.name = "Waal"
nrijn_dist = oseries.distance(nederrijn)
nrijn_dist.name = "Nederrijn"
distances = pd.concat([waal_dist, nrijn_dist], axis=1)
riv_per_pb = distances.idxmin(axis=1)

# 2a. calculate distance along riv for each piezometer
waal_linear_dist = oseries.apply(
    lambda pt: waal.project(pt.geometry), axis=1)
nrijn_linear_dist = oseries.apply(
    lambda pt: nederrijn.project(pt.geometry), axis=1)

# 2b. calculate distance to riv for each piezometer
dist_to_waal = oseries.apply(
    lambda pt: waal.distance(pt.geometry), axis=1)
dist_to_nrijn = oseries.apply(
    lambda pt: nederrijn.distance(pt.geometry), axis=1)

# 3. interpolate linear distance of piezometer between stations
wd = locs.waal_dist.sort_values().dropna()
nd = locs.nederrijn_dist.sort_values().dropna()
stn_weights_w = pd.Series(index=oseries.index,
                          data=np.interp(waal_linear_dist.values, wd.values,
                                         range(wd.index.size)))
stn_weights_n = pd.Series(index=oseries.index,
                          data=np.interp(nrijn_linear_dist.values, nd.values,
                                         range(nd.index.size)))

# 4. populate weights dataframe
weights = pd.DataFrame(index=oseries.index, columns=locs.index, data=0.0)
dist_to_riv = pd.DataFrame(index=oseries.index,
                           columns=["dist_to_nearest_riv"], data=0.0)

for ipb in oseries.index:

    if riv_per_pb.loc[ipb] == "Waal":

        w = stn_weights_w.loc[ipb]  # interpolated no.
        w_int = int(w)  # first station
        iweight = w - w_int  # interpolation weight

        idx_stn1 = wd.index[w_int]  # riv station name 1
        weights.loc[ipb, idx_stn1] = 1 - iweight

        if w_int < len(wd.index) - 1:
            idx_stn2 = wd.index[w_int + 1]  # riv station name 2
            weights.loc[ipb, idx_stn2] = iweight

        dist_to_riv.loc[ipb, "dist_to_nearest_riv"] = dist_to_waal.loc[ipb]

    elif riv_per_pb.loc[ipb] == "Nederrijn":

        w = stn_weights_n.loc[ipb]  # interpolated no.
        w_int = int(w)  # first station
        iweight = w - w_int  # interpolation weight

        idx_stn1 = nd.index[w_int]  # riv station name 1
        weights.loc[ipb, idx_stn1] = 1 - iweight

        if w_int < len(nd.index) - 1:
            idx_stn2 = nd.index[w_int + 1]  # riv station name 2
            weights.loc[ipb, idx_stn2] = iweight

        dist_to_riv.loc[ipb, "dist_to_nearest_riv"] = dist_to_nrijn.loc[ipb]

# 5. save weights
# weights.to_pickle(os.path.join(datadir, "riv_station_weights.pkl"))
# dist_to_riv.to_pickle(os.path.join(datadir, "dist_to_nearest_riv.pkl"))


# %% calculate river series per piezometer and add to store

normalize = True

for o in tqdm(conn.oseries.index, desc="Add interpolated riv series"):
    
    rivweights = weights.loc[o].nlargest(2)
    n1, n2 = rivweights.index

    # get stresses
    riv_ts_dict = conn.get_stresses(rivweights.index.to_list())

    # calculate interpolated stress
    s_riv = rivweights.loc[n1] * riv_ts_dict[n1].squeeze() + \
        rivweights.loc[n2] * riv_ts_dict[n2].squeeze()
    s_riv.name = "riv"

    if normalize:
        # s_riv_norm = (s_riv - s_riv.mean()) / s_riv.std()  # original method
        s_riv_norm = (s_riv - s_riv.mean())  # adjusted 2022-02-17
        s_riv = s_riv_norm.to_frame()
    else:
        s_riv = s_riv.to_frame()

    if riv_per_pb.loc[ipb] == "Waal":
        pass
    elif riv_per_pb.loc[ipb] == "Nederrijn":
        pass

    # NOTE: would be better to use calculated interpolated x,y position on RIV
    # as position, currently using the location of piezometer, which works but
    # not exactly what you would want for visualization purposes.
    meta = {
        "x": conn.oseries.loc[o, "x"],
        "y": conn.oseries.loc[o, "y"],
        "oseries": o,
        "normalized": normalize,
        "riv_stn1": n1,
        "riv_stn2": n2
    }

    conn.add_stress(s_riv, f"riv_interp_{o}", kind="wlvl",
                    metadata=meta, overwrite=True)
