"""Build time series models.

Run 8 model structures per location:
   1. Recharge
   2. Recharge + River
   3. Recharge + River + 1 Well
   4. Recharge + River + 2 Wells
   5. Recharge + River + 3 Wells
   6. Recharge + 1 Well
   7. Recharge + 2 Wells
   8. Recharge + 3 Wells

Wells are added to the model in the order of proximity.
"""
# %%
import json
import os
import sys

import pandas as pd
import pastas as ps
from tqdm import tqdm

sys.path.insert(1, "..")
from util import get_pastastore

ps.logger.setLevel("ERROR")

# %% setup pastastore
with open("../tsa_settings.json", "r") as f:
    settings = json.load(f)

name = f"frontiers_{settings['freq']}"
pstore = get_pastastore(name, settings["connector_type"])
conn = pstore.conn

# for speedups when storing models set timeseries checking to False
pstore.set_check_model_series_values(False)

# %% load selected models
oseries_selection = pd.read_csv(
    "../../data/piezometers/oseries_metadata_and_selection.csv", index_col=[0])
model_names = oseries_selection["selected"].dropna()

# %% split data into calibration/validation set
oseries_calib_tmin = pd.read_csv(
    "../../data/piezometers/oseries_calibration_settings.csv", index_col=[0])

# training period
tmin_train = oseries_calib_tmin["tmin_train"]
tmax_train = settings['tmax_train']

# testing period
test_start = "2015-01-01"
test_end = "2021-06-30"

# %% load well distances
distances = pd.read_csv("../../data/wells/wellfield_to_obswell_distances.csv",
                        index_col=[0])

# %% Loop through locations and create models

modelsep = settings["model_sep"]
solve = True
solver = ps.LmfitSolve
freq = "D"  # still do all the math on daily timestep
max_nfev = 5000

rsq_df = pd.DataFrame(columns=["rsq_no_noise", "rsq_w_noise"], dtype=float)

for mlname in tqdm(model_names, desc='Solve models'):

    ml0 = pstore.create_model(mlname)
    ml = ml0.copy()

    # %% Model 1: rech
    if solve:
        ml.solve(tmin=tmin_train.loc[mlname], tmax=tmax_train,
                 freq=freq, max_nfev=max_nfev,
                 solver=solver, report=False, noise=False)
        rsq0 = ml.stats.rsq()
        ml.solve(tmin=tmin_train.loc[mlname], tmax=tmax_train,
                 freq=freq, max_nfev=max_nfev,
                 solver=solver, report=False, noise=True, initial=False)
        rsqn = ml.stats.rsq()

    ml.name = mlname + f"{modelsep}100"

    if solve:
        rsq_df.loc[ml.name, "rsq_no_noise"] = rsq0
        rsq_df.loc[ml.name, "rsq_w_noise"] = rsqn

    # add model to database
    pstore.add_model(ml, overwrite=True)

    # %% Model 2: rech + riv
    s_riv_norm_ts = pstore.get_stresses(f"riv_interp_{mlname}")

    # create stressmodel
    rm = ps.StressModel(s_riv_norm_ts, ps.Exponential, "river",
                        settings="waterlevel")
    rm._set_initial("river_A", 0.5)

    # add stressmodel to model
    ml = ml0.copy()
    ml.add_stressmodel(rm)

    # solve
    if solve:
        ml.solve(tmin=tmin_train.loc[mlname], tmax=tmax_train,
                 freq=freq, max_nfev=max_nfev,
                 solver=solver, report=False, noise=False)
        rsq0 = ml.stats.rsq()
        ml.solve(tmin=tmin_train.loc[mlname], tmax=tmax_train,
                 freq=freq, max_nfev=max_nfev,
                 solver=solver, report=False, noise=True, initial=False)
        rsqn = ml.stats.rsq()
    ml.name = mlname + f"{modelsep}110"
    if solve:
        rsq_df.loc[ml.name, "rsq_no_noise"] = rsq0
        rsq_df.loc[ml.name, "rsq_w_noise"] = rsqn

    # add model to database
    pstore.add_model(ml, overwrite=True)

    # %% Models 3-8: rech + riv + wells
    # wellnames = pstore.get_nearest_stresses(mlname, kind="well", n=3)
    # dist = pstore.get_distances(
    #     mlname, kind="well").sort_values(mlname, axis=1)
    dist = distances.loc[mlname].sort_values()
    wellnames = dist.index.tolist()[:3]

    # create 3 wellmodels
    stresses = []
    for w in wellnames:
        #w = wellnames[w]
        well_series = pstore.conn.get_stresses(w)
        wmeta = pstore.conn.get_metadata("stresses", w, as_frame=False)
        ts = ps.TimeSeries(well_series, freq_original="D",
                           settings="well", metadata=wmeta, name=w)
        stresses.append(ts)

    wm1 = ps.stressmodels.WellModel(
        stresses[:1], ps.rfunc.HantushWellModel, "wells",
        distances=dist.values.squeeze()[:1].tolist(), up=False)
    wm2 = ps.stressmodels.WellModel(
        stresses[:2], ps.rfunc.HantushWellModel, "wells",
        distances=dist.values.squeeze()[:2].tolist(), up=False)
    wm3 = ps.stressmodels.WellModel(
        stresses, ps.rfunc.HantushWellModel, "wells",
        distances=dist.values.squeeze()[:3].tolist(), up=False)

    # solve with each
    for wriv in [1, 0]:
        for j, iwm in enumerate([wm1, wm2, wm3]):

            # without river
            if wriv == 0:
                ml = ml0.copy()
                prefix = "rech_"
                mlsuff = f"{modelsep}" + "10{0}"
            # with river
            else:
                ml = ml0.copy()
                ml.add_stressmodel(rm, replace=True)
                prefix = "rech_riv_"
                mlsuff = f"{modelsep}" + "11{0}"

            # add new
            ml.add_stressmodel(iwm)

            # solve
            if solve:
                ml.solve(tmin=tmin_train.loc[mlname], tmax=tmax_train,
                         freq=freq, max_nfev=max_nfev,
                         solver=solver, report=False, noise=False)
                rsq0 = ml.stats.rsq()
                ml.solve(tmin=tmin_train.loc[mlname], tmax=tmax_train,
                         freq=freq, max_nfev=max_nfev,
                         solver=solver, report=False, noise=True,
                         initial=False)
                rsqn = ml.stats.rsq()
            ml.name = mlname + mlsuff.format(j + 1)
            # track.initialize_figure(figsize=(10, 8), dpi=100)
            if solve:
                rsq_df.loc[ml.name, "rsq_no_noise"] = rsq0
                rsq_df.loc[ml.name, "rsq_w_noise"] = rsqn

            # add model to database
            pstore.add_model(ml, overwrite=True)

# %% store optimization statistics

resultsdir = f"../../results/{settings['freq']}"
if not os.path.exists(resultsdir):
    os.makedirs(resultsdir)

# store rsq after both steps in optimization
rsq_df.to_csv(os.path.join(resultsdir, "rsq_calib_all_models.csv"))
