# %%
import json
import os

import numpy as np
import pandas as pd
import pastas as ps
import pastastore as pst
from tqdm import tqdm

ps.logger.setLevel("ERROR")

# %% setup pastastore
with open("../tsa_settings.json", "r") as f:
    settings = json.load(f)

if settings["freq"] != "D":
    name = f"frontiers_{settings['freq']}"
else:
    name = "frontiers"
connstr = 'mongodb://localhost:27017/'

conn = pst.ArcticConnector(name, connstr)
pstore = pst.PastaStore(name, conn)

# for speedups set following to False
pstore.set_check_model_series_values(False)

df = pd.read_csv(
    'G:/My Drive/k/01projekt/frontiersin_special_issue/oseries - all.csv')
df = df[~df.series_sel.isna()]

# %% split data into train/test set
tmin = pstore.get_tmin_tmax('oseries').tmin
tmin_train = pd.Series(df.Year_Conclusion.values,
                          index=df.series_sel)
train_end = settings['tmax_train']  # "2014-12-31"

test_start = "2015-01-01"
test_end = "2021-06-30"

# %% Loop through locations and create models

modelsep = "||"
solve = True
solver = ps.LmfitSolve
freq = "D"  # still do all the math on daily timestep
max_nfev = 5000

# model_names = pd.Index(np.load('model_names.npy', allow_pickle=True))
model_names = pd.Index(df.series_sel)

evp_df = pd.DataFrame(columns=["evp_no_noise", "evp_w_noise"], dtype=float)

tracks = {}
for mlname in tqdm(model_names, desc='Solve Models'):
    # old - pd.to_datetime(settings['tmin_train'])
    if tmin.loc[mlname] > pd.to_datetime(tmin_train.loc[mlname]):
        train_start = str(tmin.loc[mlname].date())
    else:
        train_start = pd.to_datetime(tmin_train.loc[mlname])
    # print(mlname)
    ml0 = pstore.create_model(mlname)
    ml = ml0.copy()

    # %% Model 1: rech
    if solve:
        track = ps.TrackSolve(ml)
        ml.solve(tmin=train_start, tmax=train_end, freq=freq, max_nfev=max_nfev,
                 solver=solver, report=False, noise=False)
        evp0 = ml.stats.evp()
        ml.solve(tmin=train_start, tmax=train_end, freq=freq, max_nfev=max_nfev,
                 solver=solver, report=False, noise=True, initial=False, callback=track.track_solve)
        evpn = ml.stats.evp()
    ml.name = mlname + f"{modelsep}100"
    tracks[ml.name] = track
    if solve:
        evp_df.loc[ml.name, "evp_no_noise"] = evp0
        evp_df.loc[ml.name, "evp_w_noise"] = evpn

    # add model to database
    pstore.add_model(ml, overwrite=True)

    # %% Model 2: rech + riv
    s_riv_norm_ts = pstore.get_stresses(f"riv_interp_{mlname}")

    # create stressmodel
    rm = ps.StressModel(s_riv_norm_ts, ps.Exponential, "river",
                        settings="waterlevel")
    rm._set_initial("river_A", 0.5)

    # add stressmodel to model
    ml = ml0.copy()
    ml.add_stressmodel(rm)

    # solve
    if solve:
        ml.solve(tmin=train_start, tmax=train_end, freq=freq, max_nfev=max_nfev,
                 solver=solver, report=False, noise=False)
        evp0 = ml.stats.evp()
        ml.solve(tmin=train_start, tmax=train_end, freq=freq, max_nfev=max_nfev,
                 solver=solver, report=False, noise=True, initial=False)
        evpn = ml.stats.evp()
    ml.name = mlname + f"{modelsep}110"
    if solve:
        evp_df.loc[ml.name, "evp_no_noise"] = evp0
        evp_df.loc[ml.name, "evp_w_noise"] = evpn

    # add model to database
    pstore.add_model(ml, overwrite=True)

    # %% Models 3-8: rech + riv + wells
    wellnames = pstore.get_nearest_stresses(mlname, kind="well", n=3)
    dist = pstore.get_distances(
        mlname, kind="well").sort_values(mlname, axis=1)

    # create 3 wellmodels
    stresses = []
    for w in wellnames:
        w = wellnames.iloc[0, w]
        well_series = pstore.conn.get_stresses(w)
        wmeta = pstore.conn.get_metadata("stresses", w, as_frame=False)
        ts = ps.TimeSeries(well_series, freq_original="D",
                           settings="well", metadata=wmeta, name=w)
        stresses.append(ts)

    wm1 = ps.stressmodels.WellModel(
        stresses[:1], ps.rfunc.HantushWellModel, "wells",
        distances=dist.values.squeeze()[:1].tolist(), up=False)
    wm2 = ps.stressmodels.WellModel(
        stresses[:2], ps.rfunc.HantushWellModel, "wells",
        distances=dist.values.squeeze()[:2].tolist(), up=False)
    wm3 = ps.stressmodels.WellModel(
        stresses, ps.rfunc.HantushWellModel, "wells",
        distances=dist.values.squeeze()[:3].tolist(), up=False)

    # solve with each
    for wriv in [1, 0]:
        for j, iwm in enumerate([wm1, wm2, wm3]):

            # without river
            if wriv == 0:
                ml = ml0.copy()
                prefix = "rech_"
                mlsuff = f"{modelsep}" + "10{0}"
            # with river
            else:
                ml = ml0.copy()
                ml.add_stressmodel(rm, replace=True)
                prefix = "rech_riv_"
                mlsuff = f"{modelsep}" + "11{0}"

            # add new
            ml.add_stressmodel(iwm)

            # solve
            if solve:
                track = ps.TrackSolve(ml)
                ml.solve(tmin=train_start, tmax=train_end, freq=freq, max_nfev=max_nfev,
                         solver=solver, report=False, noise=False)
                evp0 = ml.stats.evp()
                ml.solve(tmin=train_start, tmax=train_end, freq=freq, max_nfev=max_nfev,
                         solver=solver, report=False, noise=True,
                         initial=False, callback=track.track_solve)
                evpn = ml.stats.evp()
            ml.name = mlname + mlsuff.format(j + 1)
            tracks[ml.name] = track
            # track.initialize_figure(figsize=(10, 8), dpi=100)
            if solve:
                evp_df.loc[ml.name, "evp_no_noise"] = evp0
                evp_df.loc[ml.name, "evp_w_noise"] = evpn

            # add model to database
            pstore.add_model(ml, overwrite=True)

# %%
evp_df.to_csv(f"../../results/{settings['freq']}/evp_calib_all_models.csv")
