"""Determine the best model according to different statistics.

Store final model selection as described in article. Note that the final step
is a visual check that has been hard-coded in this script to amend the final 
selection.

See script in `02_visualization/00_plot_model_results.py` for saving 
figures of model results for visual checking.
"""
# %%
import json
import os
import sys

import numpy as np
import pandas as pd
import pastas as ps
import pastastore as pst
from tqdm import tqdm

sys.path.insert(1, "..")
from util import get_pastastore

ps.logger.setLevel("ERROR")

# %%
# setup pastastore
with open("../tsa_settings.json", "r") as f:
    settings = json.load(f)

from_zip = False
if from_zip:
    fname = r'14D_tmin_1990'
    path = r'G:\.shortcut-targets-by-id\1GgFQU_fDYH7t5YumxX5Mz5XL8IgJASDp\k\01projekt\frontiersin_special_issue\02data'

    conn = pst.DictConnector('conn')
    pstore = pst.PastaStore(fname, conn)
    pstore.from_zip(f'{path}/{fname}.zip', conn)
    resultsdir = f"../../results/{fname}/modelinfo"
    csvdir = f'../../results/{fname}/'

else:
    name = f"frontiers_{settings['freq']}"
    pstore = get_pastastore(name, "pas")
    conn = pstore.conn

    resultsdir = f"../../results/{settings['freq']}/modelinfo"
    csvdir = f'../../results/{settings["freq"]}/'

if not os.path.exists(resultsdir):
    os.makedirs(resultsdir)
    print(f"Created directory: {resultsdir}")

if not os.path.exists(os.path.join(csvdir, "modelsel")):
    os.makedirs(os.path.join(csvdir, "modelsel"))
    print(f"Created directory: {os.path.join(csvdir, 'modelsel')}")


# %% get all modelinfo from csv files
res = {}
ml_res = [res[0:-4] for res in os.listdir(f'{csvdir}/modelinfo')]
for i, name in tqdm(enumerate(ml_res), desc="Read CSV"):
    res[name] = pd.read_csv(os.path.join(
        csvdir, f'modelinfo/{name}.csv'), index_col=0)

# %%
# get best models according to statistics
modelsep = "||"
mlsuffixes = ["100", "110", "111", "112", "113", "101", "102", "103"]
model_base_names = np.unique([iml.split(modelsep)[0]
                             for iml in pstore.model_names]).astype('object')

# stats to consider
stats = ['evp', 'evp_val', 'rsq', 'rsq_val', 'rsqa', 'rsqa_val',
         'nse', 'nse_val', 'rmse', 'rmse_val', 'aic', 'bic']

# store model name
bestmodel = np.empty((len(ml_res), len(stats)), dtype=object)

# store model structure
bestmodel_type = np.empty((len(ml_res), len(stats)))

# loop over all locations
for i, name in tqdm(enumerate(ml_res), desc="Get best model per stat"):
    for j in np.arange(len(stats)):
        if j < 8:
            bestmodel[i, j] = res[name].columns[np.argmax(
                res[name].loc[stats[j]].astype(float))]
        else:
            bestmodel[i, j] = res[name].columns[np.argmin(
                res[name].loc[stats[j]].astype(float))]

        bestmodel_type[i, j] = float(bestmodel[i, j][-3:])

bestmodel = pd.DataFrame(bestmodel, columns=stats, index=model_base_names)

# %% model diagnostics
d = {}

# mask 3: no autocorrelation according to runs test
mask3 = True
alpha = 0.05

# mask 1: stat > crit_value
mask1 = True
stat = 'rsq'
stat_value = 0.7

# mask 2: 2 * response length vs calibration period
mask2 = True

# mask 4: gain > gain_lim
mask4 = True
gain_lim = 0.00  # [m]

# mask 5: gain > 2x std
mask5 = True

# store AIC
aic = pd.DataFrame(np.nan, columns=mlsuffixes, index=model_base_names)

model_sel_cols = [
    "checks_passed_no",
    "checks_ml_name",
    "checks_ml_name_aic",
    f"checks_ml_name_{stat}_val"
]

sum_fails = np.zeros(np.sum([mask1, mask2, mask3, mask4, mask5]))

mls = []
for i, ml in tqdm(enumerate(model_base_names),
                  desc='Applying masks and selecting best models'):
    idf = res[ml].iloc[4:, :].astype('float')
    m = []

    if mask1:
        m1 = pd.to_numeric(idf.loc[stat]) >= stat_value
        m1.name = f'{stat} > {stat_value}'
        m.append(m1)
        sum_fails[0] += sum(~m1)

    if mask2:
        calib_period = (
            (pd.to_datetime(res[ml].loc["tmax_train"].iloc[0]) -
             pd.to_datetime(res[ml].loc["tmin_train"].iloc[0])) /
            pd.Timedelta(days=1))
        idx = ["resp_tmem_95%_recharge",
               "resp_tmem_95%_river",
               "resp_tmem_95%_wells"]
        m2 = pd.DataFrame(
            np.where(idf.loc[idx].isnull(),
                     pd.NA,
                     2 * idf.loc[idx].astype(float) <= (calib_period)),
            index=idx, columns=idf.columns)
        m.append(m2.transpose())
        sum_fails[1] += sum(~m2.all())

    if mask3:
        m3 = pd.to_numeric(idf.loc['p_value_runs_test']) > alpha
        m3.name = 'no autocorrelation runs-test'
        m.append(m3)
        sum_fails[2] += sum(~m3)

    if mask4 or mask5:
        idx_w_stderr = idf.index[
            np.where(idf.index.str.find('gain_w_stderr'), False, True)].values
        idx_std = np.concatenate(
            [['recharge_A_stderr', 'river_A_stderr'], idx_w_stderr])
        idx_w = idf.index[np.where(idf.index.str.find('gain_w_'), False, True)]
        idx_gain = np.concatenate(
            [['recharge_A', 'river_A'], idx_w[~np.isin(idx_w, idx_w_stderr)]])

    if mask4:
        m4 = pd.DataFrame(
            np.where(idf.loc[idx_gain].isnull(),
                     pd.NA,
                     np.abs(idf.loc[idx_gain].astype(float)) >= gain_lim),
            index=idx_gain, columns=idf.columns)
        m4.index = [f'{x} > {gain_lim}' for x in m4.index]
        m.append(m4.transpose())
        sum_fails[3] += sum(~m4.all())

    if mask5:
        m5 = pd.DataFrame(
            np.where(idf.loc[idx_gain].isnull(),
                     pd.NA,
                     (np.abs(idf.loc[idx_gain].reset_index(drop=True)) >=
                     2 * idf.loc[idx_std].reset_index(drop=True))),
            index=idx_gain, columns=idf.columns)
        m5.index = [f'{x} > 2*std' for x in m5.index]
        m.append(m5.transpose())
        sum_fails[4] += sum(~m5.all())
    m = pd.concat(m, axis=1).transpose()  # .astype('bool')
    checks_pass = m.sum() == m.notnull().sum()

    if checks_pass.sum() == 0:  # no models passed checks
        bestmodel.loc[ml, 'checks_passed_no'] = checks_pass.sum()
        bestmodel.loc[ml, 'checks_ml_name'] = pd.NA
        bestmodel.loc[ml, 'checks_ml_name_aic'] = pd.NA
        bestmodel.loc[ml, 'checks_ml_name_evp_val'] = pd.NA

    if checks_pass.sum() == 1:  # one model passed check
        mls.append(m.columns[checks_pass].values[0])
        bestmodel.loc[ml, 'checks_passed_no'] = checks_pass.sum()
        bestmodel.loc[ml, 'checks_ml_name'] = mls[-1]
        bestmodel.loc[ml, 'checks_ml_name_aic'] = pd.NA
        bestmodel.loc[ml, 'checks_ml_name_evp_val'] = pd.NA
        aics = pd.Series(pd.to_numeric(idf.loc['aic', checks_pass]), name=ml)
        aics.index = [i[-3:] for i in aics.index]
        aic.loc[ml] = aics

    if checks_pass.sum() > 1:  # more than one model passed checks
        bestmodel.loc[ml, 'checks_passed_no'] = checks_pass.sum()
        bestmodel.loc[ml, 'checks_ml_name'] = pd.NA
        aics = pd.Series(pd.to_numeric(idf.loc['aic', checks_pass]), name=ml)
        mls.append(aics.idxmin())
        bestmodel.loc[ml, 'checks_ml_name_aic'] = mls[-1]
        aics.index = [i[-3:] for i in aics.index]
        aic.loc[ml] = aics
        # check if we are actually looking for the max or min value
        mls.append(pd.Series(pd.to_numeric(
            idf.loc[f"{stat}_val", checks_pass])).idxmax())
        bestmodel.loc[ml, f"checks_ml_name_{stat}_val"] = mls[-1]
    m.to_csv(os.path.join(csvdir, f'modelsel/{ml}_sel.csv'))

# %% print summary
print('Total amount of models accepted '
      f'{bestmodel.checks_passed_no.sum()}')
print('Total amount of models rejected '
      f' {len(model_base_names) * 8 - bestmodel.checks_passed_no.sum()}')
print('Total amount of locations with an accepted model '
      f'{(bestmodel.checks_passed_no >= 1).sum()}')
print('Amount of locations with 0 accepted models '
      f'{(bestmodel.checks_passed_no == 0).sum()}')
print('Amount of locations with 1 accepted model '
      f'{(bestmodel.checks_passed_no == 1).sum()}')
print('Amount of locations with >1 accepted models '
      f'{(bestmodel.checks_passed_no > 1).sum()}')
print('Amount of models passed on location with >1 accepted model '
      f'{bestmodel.checks_passed_no[bestmodel.checks_passed_no > 1].sum()}')

print(len(model_base_names) * 8 - sum_fails)

# %% save CSV with best models
bestmodel.to_csv(os.path.join(csvdir, 'best_models.csv'))
