"""Parse all models and store results, grouped per location.
"""
# %%
import json
import os
import sys

import numpy as np
import pandas as pd
import pastas as ps
import pastastore as pst
from tqdm import tqdm

sys.path.insert(1, "..")
from util import get_pastastore, variance_gain

ps.logger.setLevel("ERROR")

# %% setup pastastore
with open("../tsa_settings.json", "r") as f:
    settings = json.load(f)

from_zip = False
if from_zip:
    fname = r'14D_tmin_2000'
    path = r'../../data'

    conn = pst.DictConnector('conn')
    pstore = pst.PastaStore(fname, conn)
    pstore.from_zip(f'{path}/{fname}.zip', conn)
    resultsdir = f"../../results/{fname}/modelinfo"

else:
    name = f"frontiers_{settings['freq']}"
    pstore = get_pastastore(name, settings["connector_type"])
    conn = pstore.conn

    resultsdir = f"../../results/{settings['freq']}/modelinfo"

if not os.path.exists(resultsdir):
    os.makedirs(resultsdir)
    print(f"Created directory: {resultsdir}")

# %% split data into train/test set
# train_start = settings['tmin_train']
oseries_calib_tmin = pd.read_csv(
    "../../data/piezometers/oseries_calibration_settings.csv", index_col=[0])

# training period
train_start = oseries_calib_tmin["tmin_train"]
train_end = settings['tmax_train']

# testing period
test_start = "2015-01-01"
test_end = "2021-11-01"

modelsep = settings["model_sep"]

# %%

# get all locations with models
locations = np.unique([iml.split(modelsep)[0]
                       for iml in pstore.model_names])

mlsuffixes = ["100", "110", "111", "112", "113", "101", "102", "103"]

# loop over locations
for ipb in tqdm(locations):

    # get all models
    mlnames = [modelsep.join([ipb, suff]) for suff in mlsuffixes]
    mlist = conn.get_models(mlnames, progressbar=False)

    # empty results list
    results = []

    # get number of observations and dates of train and test period
    oseries = pstore.get_oseries(ipb)
    tmin_train = mlist[0].observations().index[0].date()
    tmax_train = mlist[0].observations().index[-1].date()
    tmin_test = oseries.loc[train_start.loc[ipb]:test_end].index[0].date()
    tmax_test = oseries.loc[train_start.loc[ipb]:test_end].index[-1].date()
    nobs = len(oseries.loc[train_start.loc[ipb]:test_end])
    nobs_train = len(oseries.loc[train_start.loc[ipb]:train_end])
    nobs_test = len(oseries.loc[test_start:test_end])
    nobs_days = len(
        oseries.loc[train_start.loc[ipb]:test_end].index.normalize().unique())
    nobs_train_days = len(
        oseries.loc[train_start.loc[ipb]:train_end].index.normalize().unique())
    nobs_test_days = len(
        oseries.loc[test_start:test_end].index.normalize().unique())
    obs_list = [tmin_train, tmax_train, tmin_test, tmax_test, nobs, nobs_train,
                nobs_test, nobs_days, nobs_train_days, nobs_test_days]

    # save all info in dataframe
    obs = pd.DataFrame(
        np.transpose(np.tile(obs_list, (len(mlnames), 1))),
        columns=mlnames,
        index=pd.Index(['tmin_train', 'tmax_train', 'tmin_test', 'tmax_test',
                        'nobs', 'nobs_train', 'nobs_test', 'nobs_days',
                        'nobs_train_days', 'nobs_test_days'])).transpose()

    # get parameters, stderrs, statistics
    params = pstore.get_parameters(modelnames=mlnames, param_value="optimal")
    stderr = pstore.get_parameters(modelnames=mlnames, param_value="stderr")
    stderr.columns = [i + "_stderr" for i in stderr.columns]
    stats_calib = pstore.get_statistics(
        ["rmse", "evp", "rsq", "aic", "bic", 'nse'],
        modelnames=mlnames)

    dfr = pd.concat([obs, params, stderr, stats_calib], axis=1)

    # loop over models per location
    for ml in mlist:

        # simulate test period
        tmax_series = ml.oseries.series_original.last_valid_index()
        test_end = pd.Timestamp(test_end)

        if tmax_series > test_end:
            tmax_sim = test_end
        else:
            tmax_sim = tmax_series

        # calculated adjusted rsq
        dfr.loc[ml.name, "rsqa"] = ml.stats.rsq(
            tmin=train_start.loc[ipb], tmax=train_end,
            nparam=sum(ml.parameters.vary))

        # calculate validation stats
        dfr.loc[ml.name, "rmse_val"] = ml.stats.rmse(
            tmin=test_start, tmax=test_end)
        dfr.loc[ml.name, "evp_val"] = ml.stats.evp(
            tmin=test_start, tmax=test_end)
        dfr.loc[ml.name, "rsq_val"] = ml.stats.rsq(
            tmin=test_start, tmax=test_end)
        dfr.loc[ml.name, "nse_val"] = ml.stats.nse(
            tmin=test_start, tmax=test_end)
        dfr.loc[ml.name, "rsqa_val"] = ml.stats.rsq(
            tmin=test_start, tmax=test_end, nparam=sum(ml.parameters.vary))

        # autocorrelation
        stat, p = ps.stats.runs_test(ml.noise()[1:])
        dfr.loc[ml.name, "p_value_runs_test"] = p
        stat, p = ps.stats.stoffer_toloi(ml.noise()[1:], lags=15, freq="14D")
        dfr.loc[ml.name, "p_value_stoffer_toloi"] = p

        # get t95% memory response functions
        sm_names = ml.get_stressmodel_names()
        cutoffs = [0.95, 0.5]
        for sm_name in sm_names:
            for cutoff in cutoffs:
                # get tmax at each cutoff for response
                dfr.loc[ml.name, f"resp_tmem_{cutoff*100:0.0f}%_{sm_name}"] = \
                    ml.get_response_tmax(
                    sm_name, cutoff=cutoff)

                if sm_name == 'wells':
                    wm = ml.stressmodels['wells']
                    wm_r = wm.distances
                    wm_names = wm_r.index
                    # get parameters and store distance
                    for i in range(len(wm_names)):
                        wm_p = wm.get_parameters(model=ml)
                        dfr.loc[ml.name,
                                f"distance_r_{sm_name}_{wm_names[i]}"] = \
                            wm_r.loc[wm_names[i]]
                        if len(wm_p) == 4:
                            wm_p = [wm_p]
                        # store well response memory
                        col = (f"resp_tmem_{cutoff*100:0.0f}"
                            f"%_{sm_name}_{wm_names[i]}")
                        dfr.loc[ml.name, col] = ml.get_response_tmax(
                            sm_name, p=wm_p[i], cutoff=cutoff)
                        col = (f"resp_tmem_{50:0.0f}%_"
                            f"{sm_name}_analytic_{wm_names[i]}")
                        dfr.loc[ml.name, col] = \
                            wm_r[i] * np.sqrt(wm_p[i][2]) * wm_p[i][1]

        # calculate gain WellModel
        if "wells" in ml.stressmodels.keys():
            wm = ml.stressmodels["wells"]
            pwells = ml.get_parameters("wells")  # get base params A, a, b

            # loop over wells
            for i in range(len(wm.stress)):
                wname = wm.stress[i].name
                p_wdist = np.r_[pwells, wm.distances[i:i + 1]]

                # calculate gain
                dfr.loc[ml.name, f"gain_w_{wname}"] = wm.rfunc.gain(p_wdist)

                # calculate std using helper function
                dfr.loc[ml.name, f"gain_w_stderr_{wname}"] = \
                    np.sqrt(variance_gain(ml, 'wells', istress=i))

        # model description,
        # NOTE: turned off for now to allow setting float format in CSV
        # dfr.loc[ml.name, "description"] = \
        #     msuffix_to_description(ml.name.split(modelsep)[-1])

    # save to file
    dfr.transpose().to_csv(
        os.path.join(resultsdir, f"{ipb}.csv"), float_format="%.4g")
